# Copyright 2018 HQS Quantum Simulations
# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module provides Xmon gates for projectq.

Xmon qubits are a specific type of transmon qubit developed by Google.
In the cirq framework xmon specific gates are defined.
This file ports these gates to projectq and allows imulation
of algorithms with xmon qubits. 
"""
from projectq.ops import BasicGate, BasicPhaseGate, H, Rx, Rz, NotMergeable
import numpy as np
import cmath
from cirq import value
RTOL = 1e-10
ATOL = 1e-12
#TODO: implement merge and inverse for all gates

class XmonGate(BasicGate):
    def __init__(self):
        r"""
        Abstract XmonGate class to distinguish mon from regular proejctq gates.
        """
        BasicGate.__init__(self)

    def __eq__(self, other):
        r""" 
        Modification of ProjectQ (2017 ProjectQ-Framework (www.projectq.ch)) equality operator to deal with the transition to np.ndarray
        """
        if hasattr(self, 'matrix'):
            if not hasattr(other, 'matrix'):
                return False
        if hasattr(other, 'matrix'):
            if not hasattr(self, 'matrix'):
                return False
        if hasattr(self, 'matrix') and hasattr(other, 'matrix'):
            if (not (isinstance(self.matrix, np.matrix) or isinstance(self.matrix, np.ndarray) ) or
                    not (isinstance(other.matrix, np.matrix) or isinstance(other.matrix, np.ndarray))):
                raise TypeError("One of the gates doesn't have the correct "
                                "type (numpy.matrix) for the matrix "
                                "attribute.")
            if (self.matrix.shape == other.matrix.shape and
                np.allclose(self.matrix, other.matrix,
                            rtol=RTOL, atol=ATOL,
                            equal_nan=False)):
                return True
            else:
                return False
        else:
            return isinstance(other, self.__class__)

class Exp11Gate(XmonGate):
    r"""A two-qubit interaction that phases the amplitude of the 11 state.

    This gate implements :math:`\exp(i \pi \varphi |11\rangle\langle 11|)` where :math:`\varphi`
    is half turns.

    As a matrix it reads

    .. math::
        \begin{pmatrix}
        1 & 0 & 0 & 0\\
        0 & 1 & 0 & 0\\
        0 & 0 & 1 & 0\\
        0 & 0 & 0 & \exp(i\varphi \pi)
        \end{pmatrix}

    Warning:
        There is no (-) sign in the definition of the gate.

    Note:
        The half_turn parameter is such that a full turn is the
        identity matrix, in contrast to the single qubit gates, where a full
        turn is minus identity. The single qubit half-turn gates are defined
        so that a full turn corresponds to a rotation on the Bloch sphere of a
        360 degree rotation. For two qubit gates, there isn't a Bloch sphere,
        so the half_turn corresponds to half of a full rotation in U(4).


    Args:
        half_turns (float): angle of rotation in units of :math:`\pi`.
    """
    def __init__(self, half_turns):
        XmonGate.__init__(self)
        self.half_turns = half_turns
        self.interchangeable_qubit_indices = [[0, 1]]

    @property
    def angle(self):
        r"""Rotation angle in rad, :math:`\in(-\pi, \pi]`"""
        return self.half_turns * cmath.pi

    @property
    def half_turns(self):
        r"""Rotation angle in half turns, :math:`\in(-1, 1]`"""
        return self._half_turns

    @half_turns.setter
    def half_turns(self, half_turns):
        self._half_turns = value.chosen_angle_to_canonical_half_turns(half_turns=half_turns,
                                                                      rads=None,
                                                                      degs=None)

    @property
    def matrix(self):
        r"""Gate matrix.

        With :math:`\varphi=\mathrm{half\_turns}` the matrix implemented by the
        Exp11Gate reads as

        .. math::
            \begin{pmatrix}
            1 & 0 & 0 & 0\\
            0 & 1 & 0 & 0\\
            0 & 0 & 1 & 0\\
            0 & 0 & 0 & \exp(i\varphi \pi)
            \end{pmatrix}

        """
        return np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0],
                          [0, 0, 0, cmath.exp(1.0j * self.angle)]])

    def __str__(self):
        return "@({})".format(np.round(self.angle, 2))

    def tex_str(self):
        """
        Returns the class name and the angle as a subscript, i.e.

        .. code-block:: latex

            [CLASSNAME]$_[ANGLE]$
        """
        return "@$_{{{}}}$".format(np.round(self.angle / np.pi, 2))

class ExpWGate(XmonGate):
    r"""A rotation around an axis in the XY plane of the Bloch sphere.



    This gate is a "phased X rotation". Specifically

        ───W(axis)^t─── = ───Z^-axis───X^t───Z^axis───

    This gate is

    .. math::

        \exp(-i * \pi * W(\mathrm{axis\_half\_turn}) * \mathrm{half\_turn} / 2)

    where

    .. math::

        W(\theta) = \cos(\pi \theta) X + \sin(\pi \theta) Y

    Note the half_turn nomenclature here comes from viewing this as a rotation
    on the Bloch sphere. Two half_turns correspond to a rotation in the
    bloch sphere of 360 degrees. Note that this is minus identity, not
    just identity.  Similarly the axis_half_turns refers thinking of rotating
    the Bloch operator, starting with the operator pointing along the X
    direction. An axis_half_turn of 1 corresponds to the operator pointing
    along the -X direction while an axis_half_turn of 0.5 correspond to
    an operator pointing along the Y direction.'

    Contrary to the docstring, cirq implements the gate (:math:`\varphi = \mathrm{half\_turns}`
    and :math:`\theta=\mathrm{axis\_half\_turns}`):

    .. math::

        \exp(i \pi \varphi / 2) \exp(-i \pi W(\theta) \varphi / 2)

    which differs by a global phase :math:`\exp(i \varphi/2)`.

    This class mimics the cir implementation, i.e., the matrix represented by this gate
    reads as

    .. math::
        \exp(i \varphi \pi / 2)\begin{pmatrix}
        \cos(\varphi\pi / 2) & -i \sin(\varphi \pi / 2) exp(-i \theta \pi)\\
        -i \sin(\varphi \pi / 2) \exp(i \theta \pi) & \cos(\varphi \pi / 2)
        \end{pmatrix}

    whih is a rotation around the W-axis and an additional global phase.

    Note:
        This gate corresponds to a phase gate in the basis defined by the W-axis,
        not to a rotation gate.

    Note:
        Restricting half_turns to the rage (-1, 1] corresponds to the full range (0, 2pi]
        for a phase gate, which the ExpW gate in the cirq implementation actually is.

    Note:
        Another convention is to change a positive rotation around a negative axis to a
        negative rotation around a postive axis, i.e., half_turns -> - half_turns
        and axis_half_turns -> axis_half_turns + 1 if axis_half_turns < 0.

    Args:
        half_turns (float): angle of rotation in units of :math:`\pi`.
        axis_half_turns (float): axis between X and Y in units of :math:`\pi`.
    """
    def __init__(self, half_turns, axis_half_turns=0):
        XmonGate.__init__(self)
        self.half_turns = half_turns
        self.axis_half_turns = axis_half_turns
        if (not isinstance(self.axis_half_turns, value.Symbol) and
                not 0 <= self.axis_half_turns < 1):
            # The following code is taken from google to follow the same conventions.
            # I'm not sure if this is correct as it seems to give different matrices.
            # Canonicalize to negative rotation around positive axis.
            self.half_turns = value.canonicalize_half_turns(-self.half_turns)
            self.axis_half_turns = value.canonicalize_half_turns(
                self.axis_half_turns + 1)

    @property
    def axis_angle(self):
        r"""Axis angle in rad, :math:`\in(-\pi, \pi]`"""
        return self._axis_half_turns * cmath.pi

    @property
    def axis_half_turns(self):
        r"""Axis angle in half turns, :math:`\in(-1, 1]`"""
        return self._axis_half_turns

    @axis_half_turns.setter
    def axis_half_turns(self, half_turns):
        val = value.chosen_angle_to_canonical_half_turns(half_turns=half_turns,
                                                         rads=None,
                                                         degs=None)
        self._axis_half_turns = val

    @property
    def angle(self):
        r"""Rotation angle in rad, :math:`\in(-\pi, \pi]`"""
        return self.half_turns * cmath.pi

    @property
    def half_turns(self):
        r"""Rotation angle in half turns, :math:`\in(-1, 1]`"""
        return self._half_turns

    @half_turns.setter
    def half_turns(self, half_turns):
        val = value.chosen_angle_to_canonical_half_turns(half_turns=half_turns,
                                                         rads=None,
                                                         degs=None)
        self._half_turns = val

    @property
    def matrix(self):
        r"""Rotation matrix.

        With :math:`\varphi = \mathrm{half\_turns}` and
        :math:`\theta=\mathrm{axis\_half\_turns}` this gate implements the matrix

        .. math::

            \exp(i \varphi \pi / 2)\begin{pmatrix}
            \cos(\varphi\pi / 2) & -i \sin(\varphi \pi / 2) exp(-i \theta \pi)\\
            -i \sin(\varphi \pi / 2) \exp(i \theta \pi) & \cos(\varphi \pi / 2)
            \end{pmatrix}

        """
        W = np.exp(-1.0j * self.axis_angle)
        c = np.cos(self.angle / 2)
        s = np.sin(self.angle / 2)
        rot = np.array([[c, -1.0j * s * W],
                         [-1.0j * s * np.conj(W), c]])
        return np.exp(.5j * self.angle) * rot
        # phase = np.array([[1., 0],
        #                    [0, cmath.exp(1.0j * self.axis_angle)]])
        # c = cmath.exp(1j * self.angle)
        # rot = np.array([[1 + c, 1 - c], [1 - c, 1 + c]]) / 2
        # return phase.dot(rot).dot(np.conj(phase))

    def __str__(self):
        return "W({}, {})".format(np.round(self.angle/np.pi,2), np.round(self.axis_angle/np.pi, 2))

    def tex_str(self):
        return "W$_{{{}, {}}}$".format(np.round(self.angle/np.pi, 2),
                   np.round(self.axis_angle/np.pi, 2))

# class ExpZGate(Rz, XmonGate):
class ExpZGate(XmonGate):
    r"""A rotation around the Z axis of the Bloch sphere.

    This gate implements :math:`\exp(-i \pi Z \varphi / 2)` where Z is the Z matrix

        Z = [[1, 0], [0, -1]]

    and :math:`\varphi` are half turns.

    The full matrix reads:

    .. math::

        \begin{pmatrix}
        exp(-i \pi \varphi / 2) & 0\\
        0 & exp(i \pi \varphi / 2)
        \end{pmatrix}

    Note the half_turn nomenclature here comes from viewing this as a rotation
    on the Bloch sphere. Two half_turns correspond to a rotation in the
    bloch sphere of 360 degrees.

    Half_turns are mapped to the range (-1, 1], i.e. to rotation angles
    in the range (-pi, pi].

    Note:
        Restricting half_turns to the rage (-1, 1] corresponds to the full range (0, 2pi]
        for the phase difference between the Z eigenstates. However, we loose the global phase
        that comes with a full rotation gate in the range (0, 4pi]. Thus, the ExpZ gate is more
        like a phase then a rotation gate.

    Args:
        half_turns (float): number of half turns on the Bloch sphere.
    """

    def __init__(self, half_turns):
        # Rz.__init__(self, half_turns * cmath.pi)
        XmonGate.__init__(self)
        self.half_turns = half_turns

    @property
    def angle(self):
        r"""Rotation angle in rad, :math:`\in(-\pi, \pi]`"""
        return self.half_turns * cmath.pi

    @property
    def half_turns(self):
        r"""Rotation angle in half turns, :math:`\in(-1, 1]`"""
        return self._half_turns

    @half_turns.setter
    def half_turns(self, half_turns):
        self._half_turns = value.chosen_angle_to_canonical_half_turns(half_turns=half_turns,
                                                                      rads = None,
                                                                      degs=None)
    @property
    def matrix(self):
        r"""Rotation matrix.

        With :math:`\varphi = \mathrm{half\_turns}` this gate implements the matrix

        .. math::

            \begin{pmatrix}
            \cos(\varphi\pi / 2) - i \sin(\varphi \pi / 2) & 0\\
            0 & \cos(\varphi\pi / 2) + i \sin(\varphi \pi / 2)
            \end{pmatrix}
        """
        return np.array([[np.cos(self.angle / 2) - 1.0j * np.sin(self.angle / 2), 0],
                          [0, np.cos(self.angle / 2) + 1.0j * np.sin(self.angle / 2)]])

    def __str__(self):
        return "Z({})".format(np.round(self.angle / np.pi,2))

    def tex_str(self):
        r"""Latex representation of the gate."""
        return "Z$_{{{}}}$".format(np.round(self.angle / np.pi,2))

    def get_merged(self, other):
        """
        Return self merged with another gate.
        Default implementation handles rotation gate of the same type, where
        angles are simply added.
        Args:
            other: Rotation gate of same type.
        Raises:
            NotMergeable: For non-rotation gates or rotation gates of
                different type.
        Returns:
            New object representing the merged gates.
        """
        if isinstance(other, self.__class__):
            return self.__class__((self.angle + other.angle) / cmath.pi)
        raise NotMergeable("Can't merge different types of rotation gates.")

from projectq.backends._circuits._to_latex import get_default_settings
def drawer_settings():
    xmon_settings = get_default_settings()
    xmon_settings['gate_shadow'] = False
    xmon_settings['gates']['ExpWGate'] = {'height': 0.8, 'offset': 0.3,
                 'pre_offset': 0.2, 'width': 1.75}
    xmon_settings['gates']['ExpZGate'] = {'height': 0.8, 'offset': 0.3,
                 'pre_offset': 0.2, 'width': 1.}
    xmon_settings['gates']['Exp11Gate'] = {'height': .8, 'offset': 0.3,
                 'pre_offset': 0.3, 'width': 1.}
    return xmon_settings



# Copyright 2018 Heisenberg Quantum Simulations
# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This file provides classes to store rules for translating ProjectQ to Cirq
operations.
"""
from collections import defaultdict

class Ruleset_pq_to_cirq():
    def __init__(self, rules=[]):
        r"""
        A collection of rules to translate ProjectQ operations to Cirq
        operations.

        Args:
            rules (list of :class:`Rule_pq_to_cirq`): the rules that can be used for translations.
        """
        self._known_rules = defaultdict(list)
        self.add_rules(rules)

    def add_rules(self, rules=[]):
        r"""
        Add rules to the set of known rules.

        Args:
            rules (list of :class:`Rule_pq_to_cirq`): the rules that can be used for translations.
        """

        if hasattr(rules, '__iter__'):
            for rule in rules:
                self.add_rule(rule)
        else:
            self.add_rule(rules)

    def add_rule(self, rule):
        r"""
        Add a single rule to the set of known rules.

        Args:
            rule (:class:`Rule_pq_to_cirq`): a rule that can be used for translations.
        """
        for cls in rule.classes:
            self._known_rules[cls].append(rule.translation)

    def translate(self, cmd, mapping, qubits):
        r"""
        Translate a projectq operation into a Cirq operation.

        Args:
            cmd (:class:`projectq.ops.Command`): a projectq command instance
            mapping (:class:`dict`): a dictionary of qubit mappings
            qubits (list of :class:cirq.QubitID`): cirq qubits

        Returns:
            :class:`cirq.Operation`
        """
        if type(cmd.gate) in self._known_rules:
            return self._known_rules[type(cmd.gate)][0](cmd, mapping, qubits)
        else:
            raise TypeError

    @property
    def known_rules(self):
        """
        Dictionary of known translations.
        """
        return self._known_rules.copy()

class Rule_pq_to_cirq():
    def __init__(self, classes, translation):
        r"""
        A class to store a single translation rule from Projectq to Cirq.

        Args:
            classes (list of :class:`projectq.ops.BasicGate`): the gate classes to which the rule applies.
            translation (callable(:class:`projectq.ops.Command`, :class:`dict`, list of :class:cirq.QubitID`)): a translation to cirq
        """
        self.classes = classes
        self.translation = translation



# Copyright 2018 Heisenberg Quantum Simulations
# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""Provides proejctq engines for simulation of xmon devices.

A full engine list can be generated with :meth:`cirqproejctq.xmon_setup.xmon_engines()`
"""
from projectq import cengines, ops, setups
from . import xmon_gates, xmon_decompositions

def _filter_xmon(eng, cmd):
    '''
    Determines if a gate can be handled by xmon qubits.

    Args:
        eng(projectq.MainEngine): The engine.
        cmd(projectq.ops.Command): A projectq command object.

    Returns:
        bool: True, if cmd.gate is a valid Xmon gate.

    '''
    if isinstance(cmd.gate, ops.ClassicalInstructionGate):
        # This is required to allow Measure, Allocate, Deallocate, Flush
        return True
    elif isinstance(cmd.gate, xmon_gates.XmonGate):
        return True
    else:
        return False

def xmon_rules():
    r"""DecompositionRuleSet for decomposition into xmon gates."""
    allrules = xmon_decompositions.all_defined_decomposition_rules
    return cengines.DecompositionRuleSet(allrules)

def replacer_xmon():
    r"""Autoreplacer for decomposition into xmon gates."""
    rule_set = cengines.DecompositionRuleSet(
            modules=[xmon_decompositions, setups.decompositions])
    return cengines.AutoReplacer(rule_set)

def xmon_supported_filter():
    r"""InstructionFilter for xmon gates."""
    return cengines.InstructionFilter(_filter_xmon)

def xmon_engines():
    r"""Full engine list for simulation with xmon gates."""
    return [cengines.TagRemover(),
            cengines.LocalOptimizer(),
            replacer_xmon(),
            cengines.TagRemover(),
            cengines.LocalOptimizer(),
            xmon_supported_filter()]


# Copyright 2018 Heisenberg Quantum Simulations
# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
r"""Provides decompositon rules to decompose common gates into Xmon gates.

The module :mod:`cirqprojectq.xmon_decompositions` provides decomposition rules for
rotation gates (Rx, Ry, Rz), Pauli gates (X, Y, Z), the Hadamard gate and for CNOT
gates into native Xmon gates.
All defined rules can be imported as
:meth:`cirqprojectq.xmon_decompositions.all_defined_decomposition_rules`.

Note:
    The decompositions into xmon gates are correct up to global phases.
    If the module constant CORRECT_PHASES is set to True, ProjectQ global
    phase gates are applied to the affected qubits in order to keep track of
    the correct phase of thq qubit register.
    With :class:`projectq.setups.decompositions.ph2r` these global phase gates
    can be translated to 1-qubit phase gates (R-gate) on control qubits if
    applicable. See example below


::

    ────  C ──────       ────  C ──── R(a) ──
          |         -->       |
      ---------            ---------
    ──|       | ──       ──|       | ────────
    ──| Ph(a) | ──       ──|       | ────────
    ──|       | ──       ──|       | ────────
      ---------            ---------

A full engine list can be generated with :meth:`cirqproejctq.xmon_setup.xmon_engines()`
"""
import numpy as np
from projectq.cengines import DecompositionRule, DecompositionRuleSet
from projectq import ops, cengines, setups
from projectq.meta import Control, get_control_count
from . import xmon_gates

CORRECT_PHASES = False

def _check_phase(angle):
    return CORRECT_PHASES and (np.pi < angle <= 3 * np.pi)

all_defined_decomposition_rules = []

def _recognize_rotations(cmd):
    if isinstance(cmd.gate, (ops.Rx, ops.Ry, ops.Rz)) and get_control_count(cmd) == 0:
        return True
    else:
        return False

def _decompose_rotations(cmd):
    r"""Decompose a rotation gate into xmon gates.

    We define a rotation gate as

    .. math::

        R_i(\alpha) = \exp(-i \frac{\alpha}{2}\sigma_i)

    If :meth:`CORRECT_PHASES` is False the following decompositions will be used:

        ::

            ── Rz(a) ──  -->  ── ExpZ(a / pi) ──

        This corresponds to the following map:

        .. math::

            R_z(\alpha) \to \mathrm{ExpZ}(\alpha/\pi) =
            \begin{cases}
            -R_z(\alpha),&\pi<\alpha\leq3\pi\\
            R_z(\alpha),&\text{else}
            \end{cases}

        ::

            ── Rx(a) ──  -->  ── ExpWGate(a/pi, 1) ──

        This corresponds to the following map:

        .. math::

            R_x(\alpha) \to \mathrm{ExpW}(\alpha/\pi, 0) =
            e^{i\alpha/2}R_x(\alpha)

        ::

            ── Ry(a) ──  -->  ── ExpWGate(a/pi, 1/2) ──

        This corresponds to the following map:

        .. math::

            R_y(\alpha) \to \mathrm{ExpW}(\alpha/\pi, 1/2) =
            e^{i\alpha/2}R_y(\alpha)

    If :meth:`CORRECT_PHASES` is True, the phases are countered by a projectq phase
    gate :class:`ops.Ph` such that the obtained pahse is correct.
    """
    qb = cmd.qubits
    eng = cmd.engine
    with Control(eng, cmd.control_qubits):
        angle = cmd.gate.angle
        phase = _check_phase(angle)
        if isinstance(cmd.gate, ops.Rx):
            xmon_gates.ExpWGate(half_turns=cmd.gate.angle / np.pi, axis_half_turns=0) | qb
            if CORRECT_PHASES:
                ops.Ph(-angle/2) | qb

        elif isinstance(cmd.gate, ops.Ry):
            xmon_gates.ExpWGate(half_turns=cmd.gate.angle / np.pi, axis_half_turns=0.5) | qb
            if CORRECT_PHASES:
                ops.Ph(-angle/2) | qb
        elif isinstance(cmd.gate, ops.Rz):
            xmon_gates.ExpZGate(cmd.gate.angle / np.pi) | qb
            if phase:
                ops.Ph(np.pi) | qb


for op in [ops.Rx, ops.Ry, ops.Rz]:
    all_defined_decomposition_rules.append(DecompositionRule(op,
                                 _decompose_rotations, _recognize_rotations))

def _recognize_paulis(cmd):
    if isinstance(cmd.gate, (ops.XGate, ops.YGate, ops.ZGate)) and get_control_count(cmd) == 0:
        return True
    else:
        return False

def _decompose_paulis(cmd):
    r"""Decompose a Pauli gate into xmon gates.

    If :meth:`CORRECT_PHASES` is False the following decompositions will be used:

        ::

            Z  -->  ── ExpZ(1.0) ──

        This corresponds to the following map:

        .. math::

            Z \to \mathrm{ExpZ}(1) = e^{-i\pi/2}Z

        ::

            ── X ──  -->  ── ExpWGate(1, 0) ──

        This corresponds to the following map:

        .. math::

            X \to \mathrm{ExpW}(1, 0) = X

        ::

            ── Y ──  -->  ── ExpWGate(1, 1/2) ──

        This corresponds to the following map:

        .. math::

            Y \to \mathrm{ExpW}(1, 1/2) = Y

    If :meth:`CORRECT_PHASES` is True, the phases are countered by a projectq phase
    gate :class:`ops.Ph`.
    """
    qb = cmd.qubits
    eng = cmd.engine
    with Control(eng, cmd.control_qubits):
        if isinstance(cmd.gate, ops.XGate):
            xmon_gates.ExpWGate(half_turns=1.0, axis_half_turns=0) | qb
        elif isinstance(cmd.gate, ops.YGate):
            xmon_gates.ExpWGate(half_turns=1.0, axis_half_turns=0.5) | qb
        elif isinstance(cmd.gate, ops.ZGate):
            xmon_gates.ExpZGate(1.0) | qb
            if CORRECT_PHASES:
                ops.Ph(np.pi/2) | qb

for op in [ops.XGate, ops.YGate, ops.ZGate]:
    all_defined_decomposition_rules.append(DecompositionRule(op,
                                 _decompose_paulis, _recognize_paulis))

def _recognize_H(cmd):
    if isinstance(cmd.gate, ops.HGate) and get_control_count(cmd) == 0:
        return True
    else:
        return False

def _decompose_H(cmd):
    r"""Decompose a Hadamard gate into xmon gates.

    .. include:: <isoamsa.txt>

    If :meth:`CORRECT_PHASES` is False the following decompositions will be used:

        ::

            ── H ──  -->  ── ExpW(1/2, 1/2) ── ExpZ(1) ──

        This corresponds to the following map:

        .. math::

            H \to e^{-i\pi/4}H

    If :meth:`CORRECT_PHASES` is True, the phases are countered by a projectq phase
    gate :class:`ops.Ph`.
    """
    qb = cmd.qubits
    eng = cmd.engine
    with Control(eng, cmd.control_qubits):
        #TODO: I added a minus! Check please!
        xmon_gates.ExpWGate(half_turns=.5, axis_half_turns=.5) | qb
        xmon_gates.ExpZGate(half_turns=1.) | qb
        if CORRECT_PHASES:
            ops.Ph(np.pi/4) | qb

all_defined_decomposition_rules.append(DecompositionRule(ops.HGate,
                             _decompose_H, _recognize_H))

def _recognize_CNOT(cmd):
    if isinstance(cmd.gate, ops.XGate) and get_control_count(cmd) == 1:
        return True
    else:
        return False

def _decompose_CNOT(cmd):
    r"""Decompose a CNOT gate into two Hadamards and an Exp11Gate.

    Uses the following decomposition

    ::

        ── C ──         ───────    @     ───────
           |     -->               |
        ── X ──         ── H ── Exp11(1) ── H ──

   This corresponds to the following map:

    .. math::

        \mathrm{CNOT} \to \mathrm{CNOT}


    Warning:
        The Hadamard gates are correct Hamard gates! They have no wrong phases.
        However, in a second decomposition step these gates will each yield
        a phase of :math:`\exp(-i\pi/4)` if CORRECT_PHASES is False! In this case, the
        final map will be :math:`\mathrm{CNOT}\to\exp(-i\pi/4)\mathrm{CNOT}`

    """
    qb = cmd.qubits
    ops.H | qb
    xmon_gates.Exp11Gate(half_turns=1.0) | (cmd.control_qubits[0], qb[0])
    ops.H | qb

all_defined_decomposition_rules.append(DecompositionRule(ops.XGate,
                             _decompose_CNOT, _recognize_CNOT))


def _recognize_SWAP(cmd):
    if isinstance(cmd.gate, ops.SwapGate):
        return True
    else:
        return False

def _decompose_SWAP(cmd):
    """
    TODO
    """
    qb = cmd.qubits
    xmon_gates.ExpW(.5, .5) | qb[0]
    xmon_gates.ExpW(.5, 0) | qb[1]
    # xmon_gates.Exp11Gate(half_turns=1.0) | (cmd.control_qubits[0], qb[0])
    xmon_gates.Exp11Gate(half_turns=1.0) | (qb[0], qb[1])
    xmon_gates.ExpW(-.5, .5) | qb[0]
    xmon_gates.ExpW(-.5, 0) | qb[1]
    xmon_gates.Exp11Gate(half_turns=1.0) | (qb[0], qb[1])
    # xmon_gates.Exp11Gate(half_turns=1.0) | (cmd.control_qubits[0], qb[0])
    xmon_gates.ExpW(.5, .5) | qb[0]
    xmon_gates.ExpW(.5, 0) | qb[1]
    xmon_gates.Exp11Gate(half_turns=1.0) | (qb[0], qb[1])
    # xmon_gates.Exp11Gate(half_turns=1.0) | (cmd.control_qubits[0], qb[0])
    xmon_gates.ExpZGate(.5) | qb[0]
    xmon_gates.ExpZGate(-.5) | qb[1]

all_defined_decomposition_rules.append(DecompositionRule(ops.SwapGate,
                             _decompose_SWAP, _recognize_SWAP))



# Copyright 2018 Heisenberg Quantum Simulations
# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module provides translation rules from Projectq to Cirq for some common
gates.
"""
import cirq
version = [int(cirq.__version__[0]), int(cirq.__version__[2])]
assert(version[0] > 0 or version[1] >= 4)
import cmath
import cirq, projectq
from projectq import ops as pqo
from projectq.meta import get_control_count
from projectq.cengines import BasicEngine, DecompositionRule, DecompositionRuleSet
from cirq import ops as cop
KNOWN_SINGLE_QUBIT_GATES = {pqo.XGate: cop.X,
               pqo.YGate: cop.Y,
               pqo.ZGate: cop.Z,
               pqo.Rx: cop.XPowGate,
               pqo.Ry: cop.YPowGate,
               pqo.Rz: cop.ZPowGate,
               pqo.HGate: cop.H,
               pqo.SGate: cop.S
               }

from ._rules_pq_to_cirq import Ruleset_pq_to_cirq as Ruleset
from ._rules_pq_to_cirq import Rule_pq_to_cirq as Rule


def _rx_ry_rz(cmd, mapping, qubits):
    """
    Translate a rotation gate into a Cirq roation (phase) gate.

    Global phase difference betwee proejctq rotation gate and cirq phase gate
    is dropped.

    Args:
        cmd (:class:`projectq.ops.Command`): a projectq command instance
        mapping (:class:`dict`): a dictionary of qubit mappings
        qubits (list of :class:cirq.QubitID`): cirq qubits

    Returns:
        :class:`cirq.Operation`
    """
    gates = {pqo.Rx: cop.XPowGate,
               pqo.Ry: cop.YPowGate,
               pqo.Rz: cop.ZPowGate}
    qb_pos = [mapping[qb.id] for qr in cmd.qubits for qb in qr]
    assert len(qb_pos)==1
    cirqGate = gates[type(cmd.gate)](half_turns=cmd.gate.angle / cmath.pi)
    if get_control_count(cmd) > 0:
        ctrl_pos = [mapping[qb.id] for qb in cmd.control_qubits]
        return cop.ControlledGate(cirqGate)(*[qubits[c] for c in ctrl_pos+qb_pos])
    else:
        return cirqGate(*[qubits[idx] for idx in qb_pos])

def _pauli_gates(cmd, mapping, qubits):
    """
    Translate a Pauli (x, Y, Z) gate into a Cirq gate.

    Args:
        cmd (:class:`projectq.ops.Command`): a projectq command instance
        mapping (:class:`dict`): a dictionary of qubit mappings
        qubits (list of :class:cirq.QubitID`): cirq qubits

    Returns:
        :class:`cirq.Operation`
    """
    gates = {pqo.XGate: cop.X,
               pqo.YGate: cop.Y,
               pqo.ZGate: cop.Z}
    qb_pos = [mapping[qb.id] for qr in cmd.qubits for qb in qr]
    assert len(qb_pos)==1
    cirqGate = gates[type(cmd.gate)]
    if get_control_count(cmd) > 0:
        ctrl_pos = [mapping[qb.id] for qb in cmd.control_qubits]
        return cop.ControlledGate(cirqGate)(*[qubits[c] for c in ctrl_pos+qb_pos])
    else:
        return cirqGate(*[qubits[idx] for idx in qb_pos])

def _h_s_gate(cmd, mapping, qubits):
    """
    Translate a Hadamard or S-gate into a Cirq gate.

    Args:
        cmd (:class:`projectq.ops.Command`): a projectq command instance
        mapping (:class:`dict`): a dictionary of qubit mappings
        qubits (list of :class:cirq.QubitID`): cirq qubits

    Returns:
        :class:`cirq.Operation`
    """
    gates = {pqo.HGate: cop.H,
               pqo.SGate: cop.S}
    qb_pos = [mapping[qb.id] for qr in cmd.qubits for qb in qr]
    assert len(qb_pos)==1
    cirqGate = gates[type(cmd.gate)]
    if get_control_count(cmd) > 0:
        ctrl_pos = [mapping[qb.id] for qb in cmd.control_qubits]
        return cop.ControlledGate(cirqGate)(*[qubits[c] for c in ctrl_pos+qb_pos])
    else:
        return cirqGate(*[qubits[idx] for idx in qb_pos])

def _gates_with_known_matrix(cmd, mapping, qubits):
    """
    Translate a single qubit gate with known matrix into a Cirq gate.

    Args:
        cmd (:class:`projectq.ops.Command`): a projectq command instance
        mapping (:class:`dict`): a dictionary of qubit mappings
        qubits (list of :class:cirq.QubitID`): cirq qubits

    Returns:
        :class:`cirq.Operation`
    """
    gate = cmd.gate
    if len(get_control_count(cmd)) == 0 and hasattr(gate, 'matrix'):
            qb_pos = [mapping[qb.id] for qr in cmd.qubits for qb in qr]
            return cop.matrix_gates.SingleQubitMatrixGate(matrix=gate.matrix)(
                    *[qubits[q] for q in qb_pos])
    elif len(get_control_count(cmd)) > 0 and hasattr(gate, 'matrix'):
        qb_pos = [mapping[qb.id] for qr in cmd.qubits for qb in qr]
        ctrl_pos = [mapping[qb.id] for qb in cmd.control_qubits]
        cirqGate = cop.matrix_gates.SingleQubitMatrixGate(matrix=gate.matrix)
        return cop.ControlledGate(cirqGate)(*[qubits[q] for q in qb_pos+ctrl_pos])


Rx_Ry_Rz = Rule([pqo.Rx, pqo.Ry, pqo.Rz], _rx_ry_rz)
Paulis = Rule([pqo.XGate, pqo.YGate, pqo.ZGate], _pauli_gates)
H_S = Rule([pqo.HGate, pqo.SGate], _h_s_gate)
Known_Matrix = Rule([pqo.BasicGate], _gates_with_known_matrix)

common_gates_ruleset = Ruleset(rules = [Rx_Ry_Rz, Paulis, H_S, Known_Matrix])


import pytest
import numpy as np
import itertools
from numpy import testing as nptest
from cirqprojectq import xmon_gates
import cirq
from cirq import ops
version = [int(cirq.__version__[0]), int(cirq.__version__[2])]
if (version[0] == 0 and version[1] <= 3):
    from cirq import google
else:
    from cirq.protocols.unitary import unitary

@pytest.mark.parametrize("half_turns", [-1.5, -1., -.5, 0, 0.25, 0.3, 1., 1.5, 2.])
def test_expz_matrix(half_turns):
    if (version[0] == 0 and version[1] <= 3):
        nptest.assert_array_almost_equal(xmon_gates.ExpZGate(half_turns=half_turns).matrix,
                                google.ExpZGate(half_turns=half_turns).matrix())
    else:
        nptest.assert_array_almost_equal(xmon_gates.ExpZGate(half_turns=half_turns).matrix,
                                unitary(ops.ZPowGate(exponent=half_turns,
                                                     global_shift=half_turns)))


@pytest.mark.parametrize("half_turns, axis_half_turns",
                         list(itertools.product([-1.5, -1., -.5, 0, 0.25, 0.3, 1., 1.5, 2.],
                                                [-1.5, -1., -.5, 0, 0.25, 0.3, 1., 1.5, 2.])))
def test_expw_matrix(half_turns, axis_half_turns):
    m1 = xmon_gates.ExpWGate(half_turns=half_turns,
                                axis_half_turns=axis_half_turns).matrix
    if (version[0] == 0 and version[1] <= 3):
        m2 = google.ExpWGate(half_turns=half_turns,
                            axis_half_turns=axis_half_turns).matrix()
    else:
         m2 = unitary(ops.PhasedXPowGate(exponent=half_turns,
                            phase_exponent=axis_half_turns))
    nptest.assert_array_almost_equal(m1, m2)


@pytest.mark.parametrize("half_turns", [-1.5, -1., -.5, 0, 0.25, 0.3, 1., 1.5, 2.])
def test_expw_matrix(half_turns):
    if (version[0] == 0 and version[1] <= 3):
        nptest.assert_array_almost_equal(xmon_gates.Exp11Gate(half_turns=half_turns).matrix,
                                google.Exp11Gate(half_turns=half_turns).matrix())
    else:
        nptest.assert_array_almost_equal(xmon_gates.Exp11Gate(half_turns=half_turns).matrix,
                                unitary(ops.CZPowGate(exponent=half_turns)))


# Copyright 2018 Heisenberg Quantum Simulations
# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""This package provdes a port between Cirq and ProejctQ.

Cirq is an open source quantum computing framework developed by Google and
Projectq is a framework to compile and simulate quantum algorithms.

Cirqprojectq provides two main functionalities: Firstly, it provides methods and
classes that can port algorithms from projectq to cirq and secondly it introduces
Xmon gates and xmon gate compositions to projectq.

.. autosummary::

    xmon_gates
    xmon_decompositions
    xmon_rules
    xmon_setup
    circ_engine
"""

from . import (xmon_gates,
               xmon_decompositions,
               xmon_rules,
               xmon_setup,
               circ_engine)


from setuptools import find_packages, setup

with open('README.md') as file:
    readme = file.read()

with open('LICENSE') as file:
    license = file.read()

install_requires = [
		'cirq',
		'projectq',
    		'numpy',
    		'pytest',
]

setup(name='CirqProjectQ',
      description=('Provides: ProjectQ decompositions from common gates to native cirq Xmon gates;'
                   +'  ProjectQ backend to convert a ProjectQ algorithm to a cirq.Circuit'),
      version='0.2.2',
      long_description=readme,
      packages=find_packages(exclude=('docs')),
      author='HQS Quantum Simulations',
      author_email='info@quantumsimulations.de',
      url='https://github.com/HQSquantumsimulations/CirqProjectQ',
      download_url='https://github.com/HQSquantumsimulations/CirqProjectQ/archive/v0.2.1.tar.gz',
      license=license,
      install_requires=install_requires
      )



# Copyright 2018 Heisenberg Quantum Simulations
# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module provides translation rules from Xmon gates in Projectq to Xmon gates
in Cirq.
"""
import cirq
version = [int(cirq.__version__[0]), int(cirq.__version__[2])]
assert(version[0] > 0 or version[1] >= 4)
import cmath
from projectq.meta import get_control_count
from . import xmon_gates
from ._rules_pq_to_cirq import Ruleset_pq_to_cirq as Ruleset
from ._rules_pq_to_cirq import Rule_pq_to_cirq as Rule
#from cirq.google import xmon_gates as cxmon
from cirq import ops as cop
def _expWGate(cmd, mapping, qubits):
    """
    Translate a ExpW gate into a Cirq gate.

    Args:
        - cmd (:class:`projectq.ops.Command`) - a projectq command instance
        - mapping (:class:`dict`) - a dictionary of qubit mappings
        - qubits (list of :class:cirq.QubitID`) - cirq qubits

    Returns:
        - :class:`cirq.Operation`
    """
    qb_pos = [mapping[qb.id] for qr in cmd.qubits for qb in qr]
    assert len(qb_pos)==1
    cirqGate = cop.PhasedXPowGate(exponent=cmd.gate.angle / cmath.pi, phase_exponent=cmd.gate.axis_angle / cmath.pi)
    if get_control_count(cmd) > 0:
        ctrl_pos = [mapping[qb.id] for qb in cmd.control_qubits]
        return cop.ControlledGate(cirqGate)(*[qubits[c] for c in ctrl_pos+qb_pos])
    else:
        return cirqGate(*[qubits[idx] for idx in qb_pos])

def _expZGate(cmd, mapping, qubits):
    """
    Translate a ExpZ gate into a Cirq gate.

    Args:
        - cmd (:class:`projectq.ops.Command`) - a projectq command instance
        - mapping (:class:`dict`) - a dictionary of qubit mappings
        - qubits (list of :class:cirq.QubitID`) - cirq qubits

    Returns:
        - :class:`cirq.Operation`
    """
    qb_pos = [mapping[qb.id] for qr in cmd.qubits for qb in qr]
    assert len(qb_pos)==1
    cirqGate = cop.ZPowGate(exponent=cmd.gate.angle / cmath.pi,
                            global_shift=cmd.gate.angle / cmath.pi)
    if get_control_count(cmd) > 0:
        ctrl_pos = [mapping[qb.id] for qb in cmd.control_qubits]
        return cop.ControlledGate(cirqGate)(*[qubits[c] for c in ctrl_pos+qb_pos])
    else:
        return cirqGate(*[qubits[idx] for idx in qb_pos])

def _exp11Gate(cmd, mapping, qubits):
    """
    Translate a ExpW gate into a Cirq gate.

    Args:
        - cmd (:class:`projectq.ops.Command`) - a projectq command instance
        - mapping (:class:`dict`) - a dictionary of qubit mappings
        - qubits (list of :class:cirq.QubitID`) - cirq qubits

    Returns:
        - :class:`cirq.Operation`
    """
    qb_pos = [mapping[qb.id] for qr in cmd.qubits for qb in qr]
    assert len(qb_pos)==2
    cirqGate = cop.CZPowGate(exponent=cmd.gate.angle / cmath.pi)
    return cirqGate(*[qubits[idx] for idx in qb_pos])

EXP_W = Rule([xmon_gates.ExpWGate], _expWGate)
EXP_Z = Rule([xmon_gates.ExpZGate], _expZGate)
EXP_11 = Rule([xmon_gates.Exp11Gate], _exp11Gate)
ALL_RULES = [EXP_W, EXP_Z, EXP_11]
xmon_gates_ruleset = Ruleset(rules = [EXP_W, EXP_Z, EXP_11])


# Copyright 2018 Heisenberg Quantum Simulations
# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Provides a projectq engine that translates a projectq circuit to a cirq circuit.
"""
import cirq
version = [int(cirq.__version__[0]), int(cirq.__version__[2])]
if (version[0] == 0 and version[1] <= 3):
    from .common_rules_03x import common_gates_ruleset
else:
    from .common_rules_040 import common_gates_ruleset
from projectq import ops as pqo
from projectq.cengines import BasicEngine
from projectq.meta import get_control_count
from . import xmon_rules
import cirq

class CIRQ(BasicEngine):
    r"""
    A projectq backend designated to translating to cirq.

    Args:
        qubits (list(:class:`cirq.devices.grid_qubit`)): the qubits
        device (:class:`cirq.devices.Device`): a device that provides the qubits.
        rules (cirqprojectq._rules_pq_to_cirq.Ruleset_pq_to_cirq): rule set
        strategy (:class:`cirq.circuits.InsertStrategy`): Insert strategy in cirq.
    """
    def __init__(self, qubits=None, device=None, rules=None,
                 strategy=cirq.circuits.InsertStrategy.EARLIEST):
        BasicEngine.__init__(self)
        self.strategy = strategy
        #TODO: Make rules an input.
        if rules is None:
            self._rules = common_gates_ruleset
            self._rules.add_rules(xmon_rules.ALL_RULES)

        assert not (qubits is None and device is None), "Please specify one of qubits or device!"
        self._device = device
        self._qubits = qubits or device.qubits
        self._reset()
        self._new = True

    @property
    def circuit(self):
        r"""
        :class:`cirq.Circuit`: the circuit stored in the engine.
        """
        return self._circuit

# TODO: use a device?
    @property
    def device(self):
        r"""
        :class:`cirq.devices.Device`: A device. Currently not used.
        """
        return self._device

    @property
    def qubits(self):
        r"""
        list(:class:`cirq.QubitID`): The cirq qubits used in the circuit.
        """
        return self._qubits

    def _reset(self):
        r"""Resets the circuit."""
        self._circuit = cirq.circuits.Circuit()
        self._operations = []
        self._mapping = dict()
        self._inverse_mapping = dict()

    def reset(self, keep_map=True):
        r"""Resets the engine."""
        map_ = self._mapping.copy()
        imap_ = self._inverse_mapping.copy()
        self._reset()
        self._new = True
        if keep_map:
            self._mapping = map_
            self._inverse_mapping = imap_

    def is_available(self, cmd):
        """
        Returns true if the command can be translated.

        Args:
            cmd (Command): Command for which to check availability
        """
        if cmd.gate in (pqo.Measure, pqo.Allocate, pqo.Deallocate, pqo.Barrier):
            return True
        else:
            if type(cmd.gate) in self._rules.known_rules:
                return True
            else:
                return False


    def _store(self, cmd):
        r"""Append operations given in cmd to the :class:`cirq.Circuit`.

        Args:
            cmd: Projectq command.
        """
        if self._new:
            self._new = False
            self._operations = []
        if cmd.gate == pqo.Allocate:
            qb_id = cmd.qubits[0][0].id
            #TODO placement
#            for tag in cmd.tags:
#                if isinstance(tag, QubitPlacementTag):
#                    self._mapping[qb_id] = tag.position
#                    self._inverse_mapping[tag.position] = qb_id
#                    break
            if qb_id not in self._mapping:
                self._mapping[qb_id] = qb_id
#                raise Exception("No qubit placement info found in Allocate.\n"
#                                "Please make sure you are using the CIRQ Mapper")
            # TODO check if id in device.qubits
            return
        elif cmd.gate in (pqo.Allocate, pqo.Deallocate, pqo.Barrier):
            return
        else:
            try:
                self._operations.append(self._rules.translate(cmd, self._mapping, self._qubits))
            except:
                raise TypeError("Gate {} not known".format(cmd.gate.__class__))

    def _run(self):
        r"""Appends operations to circuit and resets operations."""
        self.circuit.append(self._operations,
                            strategy=cirq.circuits.InsertStrategy.EARLIEST)
        self._operations = []

    def receive(self, command_list):
        """
        Receives a command list and, for each command, stores it until
        completion.

        Args:
            command_list: List of commands to execute
        """
        for cmd in command_list:
            if not cmd.gate == pqo.FlushGate():
                self._store(cmd)
            else:
                self._run()
                self._new = True




# Copyright 2018 Heisenberg Quantum Simulations
# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module provides translation rules from Projectq to Cirq for some common
gates.
"""
import cirq
version = [int(cirq.__version__[0]), int(cirq.__version__[2])]
if (version[0] == 0 and version[1] <= 3):
    from . import common_rules_03x
else:
    from . import common_rules_040